from binascii import hexlify

from luqum.parser import parser
from luqum.utils import LuceneTreeVisitorV2

from sqlalchemy.orm.attributes import InstrumentedAttribute

from model import *
from sqlalchemy.ext import baked
from sqlalchemy import and_, or_, not_


class SQLQueryBuilderBaseException(Exception):
    """
        Base
    """


class UnsupportedGrammarException(SQLQueryBuilderBaseException):
    """
        Raised when SQLQueryBuilder does not support given lucene grammar
    """


class FieldNotQueryableException(SQLQueryBuilderBaseException):
    """
        Raised when field does not exists, so it can't be queried, eg. something.unexistent_field
    """


class MultipleObjectsQueryException(SQLQueryBuilderBaseException):
    """
        Raised when multiple objects are queried
    """


class SQLQueryBuilder(LuceneTreeVisitorV2):
    def __init__(self):
        bakery = baked.bakery()
        self.baked_query = bakery(lambda session: session.query(BadGuy))

        self.queried_class = None
        self.wildcard_map = {
            "*": "%",
            "?": "_"
        }

    def visit_and_operation(self, node, parents, context):
        child_nodes = [self.visit(child_node, parents=parents + [node], context=context) for child_node in
                       node.children]

        conditions = and_(*child_nodes)
        return conditions

    def visit_or_operation(self, node, parents, context):
        child_nodes = [self.visit(child_node, parents=parents + [node], context=context) for child_node in
                       node.children]

        conditions = or_(*child_nodes)
        return conditions

    def visit_search_field(self, node, parents, context):
        object, column = self._get_object_column(node.name)

        # Leave this for debug
        if type(column) is not InstrumentedAttribute:
            column = hexlify(str(column).encode('utf-8')).decode('ascii')

        context["column"] = column

        condition = self.visit(node.expr, parents=parents + [node], context=context)

        return condition

    def visit_not(self, node, parents, context):
        # .a is NOT's child
        child_node = node.a

        condition = not_(self.visit(child_node, parents=parents, context=context))
        return condition

    def visit_prohibit(self, node, parents, context):
        child_node = node.children[0]

        condition = not_(self.visit(child_node, parents=parents, context=context))
        return condition

    def visit_word(self, node, parents, context):
        # Word("asd")
        column = context["column"]

        condition = self._get_condition(column, node)

        return condition

    def visit_phrase(self, node, parents, context):
        # Phrase("asd")
        column = context["column"]

        # Strip the " from start and end
        node.value = node.value[1:-1]
        condition = self._get_condition(column, node)

        return condition

    def _get_condition(self, column, node):
        value = node.value

        # No wildcard support for now
        if node.has_wildcard():
            raise NotImplementedError("No wildcard support")

        condition = column == value

        return condition

    def visit_range(self, node, parents, context):
        # [0 TO 10000] {0 TO 10000}
        column = context["column"]

        condition = and_(column > node.low.value)
        condition = and_(condition, column < node.high.value)
        if node.include_low:
            condition = or_(condition, column == node.low.value)

        if node.include_high:
            condition = or_(condition, column == node.high.value)

        return condition

    def visit_group(self, node, parents, context):
        condition = self.visit(node.expr, parents=parents, context=context)
        return condition

    def visit_field_group(self, node, parents, context):
        condition = self.visit(node.expr, parents=parents, context=context)
        return condition

    def _get_object_column(self, field_name):
        field = field_name.split(".")

        if len(field) == 0:
            raise FieldNotQueryableException("Fail")

        if len(field) == 1 or field[0] != "badguy":
            field.insert(0, "badguy")

        mapping = {
            "badguy": BadGuy,
        }

        object_type_name = field[0]
        object_type_class = mapping.get(object_type_name, None)

        if object_type_class is None:
            raise FieldNotQueryableException("There is no model named {}".format(object_type_name))

        column = self._get_column_by_name(object_type_class, field[1:])
        return object_type_class, column

    def _get_column_by_name(self, object, field):
        column = object

        for f in field:
            column = getattr(column, f, None)
            if column is None:
                raise FieldNotQueryableException("There is not field named {}".format(f))
        return column

    def visit_unknown_operation(self, node, parents, context):
        raise UnsupportedGrammarException("{} is not supported".format(str(node)))

    def __call__(self, tree):
        condition = self.visit(tree, context={})
        self.baked_query += lambda q: q.filter(condition)
        return self.baked_query


def search(query):
    builder = SQLQueryBuilder()
    tree = parser.parse(query)
    baked_query = builder(tree)
    return baked_query
